# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel schema validation model
"""

from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from itertools import chain
from math import isnan, isinf
from uuid import UUID


# Validation mode
class ValidationMode(Enum):
    """
    TODO
    """

    #: TODO
    DEFAULT = 0

    #: TODO
    QUERY_STRING = 1

    #: TODO
    JSON_INPUT = 2


# Immutable validation modes
IMMUTABLE_VALIDATION_MODES = (ValidationMode.DEFAULT,)


class AttributeValidationError(Exception):
    """
    TODO

    :param str attr: TODO
    """

    __slots__ = ('attr',)

    def __init__(self, attr):
        super().__init__(f"Invalid attribute '{attr}'")

        #: TODO
        self.attr = attr


class ValidationError(Exception):
    """
    TODO

    :param str msg: TODO
    :param str member: TODO
    """

    __slots__ = ('member',)

    def __init__(self, msg, member=None):
        super().__init__(msg)

        #: TODO
        self.member = member


def _flatten_members(members):
    for submember in members:
        if isinstance(submember, tuple):
            yield from _flatten_members(submember)
        else:
            yield submember

def _member_syntax(members):
    if members:
        return ''.join(((f'.{x}') if isinstance(x, str) else (f'[{x!r}]')) for x in _flatten_members(members)).lstrip('.')
    return None

def _member_error(type_, value, members, constraint_syntax=None):
    member_syntax = _member_syntax(members)
    member_part = f" for member '{member_syntax}'" if member_syntax else ''
    type_part = f", expected type '{type_.type_name}'" if type_ else ''
    constraint_part = f' [{constraint_syntax}]' if constraint_syntax else ''
    msg = f"Invalid value {value!r:.1000s} (type '{value.__class__.__name__}'){member_part}{type_part}{constraint_part}"
    return ValidationError(msg, member=member_syntax)


def get_referenced_types(type_):
    """
    TODO

    :param type_: TODO
    :type type_: ~chisel.TypeStruct or ~chisel.TypeEnum or ~chisel.Typedef or ~chisel.TypeArray or ~chisel.TypeDict or ~chisel.ActionModel
    """

    return sorted(_get_referenced_types(type_, set(), True), key=lambda type_: type_.type_name)

def _get_referenced_types(type_, visited, top_level):
    if isinstance(type_, TypeStruct):
        if type_.type_name not in visited:
            visited.add(type_.type_name)
            if not top_level:
                yield type_
            for member in type_.members():
                yield from _get_referenced_types(member.type, visited, False)
    elif isinstance(type_, TypeEnum):
        if type_.type_name not in visited:
            visited.add(type_.type_name)
            if not top_level:
                yield type_
    elif isinstance(type_, Typedef):
        if type_.type_name not in visited:
            visited.add(type_.type_name)
            if not top_level:
                yield type_
            yield from _get_referenced_types(type_.type, visited, False)
    elif isinstance(type_, TypeArray):
        yield from _get_referenced_types(type_.type, visited, False)
    elif isinstance(type_, TypeDict):
        yield from _get_referenced_types(type_.type, visited, False)
        yield from _get_referenced_types(type_.key_type, visited, False)
    elif isinstance(type_, ActionModel):
        yield from _get_referenced_types(type_.path_type, visited, True)
        yield from _get_referenced_types(type_.query_type, visited, True)
        yield from _get_referenced_types(type_.input_type, visited, True)
        yield from _get_referenced_types(type_.output_type, visited, True)


class ActionModel:
    """
    TODO

    :param str name: TODO
    :param str doc: The documentation markdown text
    :param str doc_group: TODO
    """

    __slots__ = ('name', 'urls', 'path_type', 'query_type', 'input_type', 'output_type', 'error_type', 'doc', 'doc_group')

    def __init__(self, name, doc=None, doc_group=None):

        #: TODO
        self.name = name

        #: TODO
        self.urls = []

        #: TODO
        self.path_type = TypeStruct(type_name=name + '_path')

        #: TODO
        self.query_type = TypeStruct(type_name=name + '_query')

        #: TODO
        self.input_type = TypeStruct(type_name=name + '_input')

        #: TODO
        self.output_type = TypeStruct(type_name=name + '_output')

        #: TODO
        self.error_type = TypeEnum(type_name=name + '_error')

        #: The documentation markdown text
        self.doc = doc

        #: TODO
        self.doc_group = doc_group

    def input_members(self, include_base_types=True):
        """
        TODO

        :param bool include_base_types: TODO
        """

        yield from chain.from_iterable(
            t.members(include_base_types=include_base_types) for t in (self.path_type, self.query_type, self.input_type)
        )


# Struct member attributes
class StructMemberAttributes:
    """
    TODO

    :param float op_eq: TODO
    :param float op_lt: TODO
    :param float op_lte: TODO
    :param float op_gt: TODO
    :param float op_gte: TODO
    :param float op_len_eq: TODO
    :param float op_len_lt: TODO
    :param float op_len_lte: TODO
    :param float op_len_gt: TODO
    :param float op_len_gte: TODO
    """

    __slots__ = ('op_eq', 'op_lt', 'op_lte', 'op_gt', 'op_gte',
                 'op_len_eq', 'op_len_lt', 'op_len_lte', 'op_len_gt', 'op_len_gte')

    def __init__(self, op_eq=None, op_lt=None, op_lte=None, op_gt=None, op_gte=None,
                 op_len_eq=None, op_len_lt=None, op_len_lte=None, op_len_gt=None, op_len_gte=None):

        #: TODO
        self.op_eq = op_eq

        #: TODO
        self.op_lt = op_lt

        #: TODO
        self.op_lte = op_lte

        #: TODO
        self.op_gt = op_gt

        #: TODO
        self.op_gte = op_gte

        #: TODO
        self.op_len_eq = op_len_eq

        #: TODO
        self.op_len_lt = op_len_lt

        #: TODO
        self.op_len_lte = op_len_lte

        #: TODO
        self.op_len_gt = op_len_gt

        #: TODO
        self.op_len_gte = op_len_gte

    @staticmethod
    def _format_float(value):
        return f'{value:.6f}'.rstrip('0').rstrip('.')

    def validate(self, value, _member=()):
        """
        TODO

        :param object value: TODO
        """

        if self.op_lt is not None and value >= self.op_lt:
            raise _member_error(None, value, _member, constraint_syntax='< ' + self._format_float(self.op_lt))
        if self.op_lte is not None and value > self.op_lte:
            raise _member_error(None, value, _member, constraint_syntax='<= ' + self._format_float(self.op_lte))
        if self.op_gt is not None and value <= self.op_gt:
            raise _member_error(None, value, _member, constraint_syntax='> ' + self._format_float(self.op_gt))
        if self.op_gte is not None and value < self.op_gte:
            raise _member_error(None, value, _member, constraint_syntax='>= ' + self._format_float(self.op_gte))
        if self.op_eq is not None and value != self.op_eq:
            raise _member_error(None, value, _member, constraint_syntax='== ' + self._format_float(self.op_eq))
        if self.op_len_lt is not None and len(value) >= self.op_len_lt:
            raise _member_error(None, value, _member, constraint_syntax='len < ' + self._format_float(self.op_len_lt))
        if self.op_len_lte is not None and len(value) > self.op_len_lte:
            raise _member_error(None, value, _member, constraint_syntax='len <= ' + self._format_float(self.op_len_lte))
        if self.op_len_gt is not None and len(value) <= self.op_len_gt:
            raise _member_error(None, value, _member, constraint_syntax='len > ' + self._format_float(self.op_len_gt))
        if self.op_len_gte is not None and len(value) < self.op_len_gte:
            raise _member_error(None, value, _member, constraint_syntax='len >= ' + self._format_float(self.op_len_gte))
        if self.op_len_eq is not None and len(value) != self.op_len_eq:
            raise _member_error(None, value, _member, constraint_syntax='len == ' + self._format_float(self.op_len_eq))

    def validate_attr(self, allow_value=False, allow_length=False):
        """
        TODO

        :param bool allow_value: TODO
        :param bool allow_length: TODO
        """

        if not allow_value:
            if self.op_lt is not None:
                raise AttributeValidationError('< ' + self._format_float(self.op_lt))
            if self.op_lte is not None:
                raise AttributeValidationError('<= ' + self._format_float(self.op_lte))
            if self.op_gt is not None:
                raise AttributeValidationError('> ' + self._format_float(self.op_gt))
            if self.op_gte is not None:
                raise AttributeValidationError('>= ' + self._format_float(self.op_gte))
            if self.op_eq is not None:
                raise AttributeValidationError('== ' + self._format_float(self.op_eq))
        if not allow_length:
            if self.op_len_lt is not None:
                raise AttributeValidationError('len < ' + self._format_float(self.op_len_lt))
            if self.op_len_lte is not None:
                raise AttributeValidationError('len <= ' + self._format_float(self.op_len_lte))
            if self.op_len_gt is not None:
                raise AttributeValidationError('len > ' + self._format_float(self.op_len_gt))
            if self.op_len_gte is not None:
                raise AttributeValidationError('len >= ' + self._format_float(self.op_len_gte))
            if self.op_len_eq is not None:
                raise AttributeValidationError('len == ' + self._format_float(self.op_len_eq))


class Typedef:
    """
    TODO

    :param object type_: TODO
    :param ~chisel.StructMemberAttributes attr: TODO
    :param str type_name: TODO
    :param str doc: The documentation markdown text
    """

    __slots__ = ('type_name', 'type', 'attr', 'doc')

    def __init__(self, type_, attr=None, type_name=None, doc=None):

        #: TODO
        self.type_name = 'typedef' if type_name is None else type_name

        #: TODO
        self.type = type_

        #: TODO
        self.attr = attr

        #: The documentation markdown text
        self.doc = doc

    @staticmethod
    def base_type(type_):
        """
        TODO
        """

        while isinstance(type_, Typedef):
            type_ = type_.type
        return type_

    def validate_attr(self, attr):
        """
        TODO
        """

        self.type.validate_attr(attr)

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        result = self.type.validate(value, mode, _member)
        if self.attr is not None:
            self.attr.validate(result, _member)
        return result


class StructMember:
    """
    TODO

    :param str name: TODO
    :param object type_: TODO
    :param bool optional: TODO
    :param bool nullable: TODO
    :param ~chisel.StructMemberAttributes attr: TODO
    :param str doc: The documentation markdown text
    """

    __slots__ = ('name', 'type', 'optional', 'nullable', 'attr', 'doc')

    def __init__(self, name, type_, optional=False, nullable=False, attr=None, doc=None):

        #: TODO
        self.name = name

        #: TODO
        self.type = type_

        #: TODO
        self.optional = optional

        #: TODO
        self.nullable = nullable

        #: TODO
        self.attr = attr

        #: The documentation markdown text
        self.doc = doc


class TypeStruct:
    """
    TODO

    :param str type_name: The struct type's name
    :param bool union: TODO
    :param list(~chisel.TypeStruct) base_types: TODO
    :param str doc: The documentation markdown text
    """

    __slots__ = ('type_name', 'union', 'base_types', '_members', 'doc')

    def __init__(self, type_name=None, union=False, base_types=None, doc=None):

        #: TODO
        self.type_name = ('union' if union else 'struct') if type_name is None else type_name

        #: TODO
        self.union = union

        #: TODO
        self.base_types = base_types

        self._members = []

        #: The documentation markdown text
        self.doc = doc

    def members(self, include_base_types=True):
        """
        Get an iterator of a struct's members

        :param bool include_base_types: If true, members from any base classes are included.
        :returns: Iterator of the struct members
        """

        if include_base_types and self.base_types is not None:
            return chain(
                chain.from_iterable(Typedef.base_type(base_type).members() for base_type in self.base_types if base_type),
                self._members
            )
        return iter(self._members)

    def add_member(self, name, type_, optional=False, nullable=False, attr=None, doc=None):
        """
        TODO

        :param str name: TODO
        :param object type_: TODO
        :param bool optional: TODO
        :param bool nullable: TODO
        :param ~chisel.StructMemberAttributes attr: TODO
        :param str doc: The documentation markdown text
        """

        member = StructMember(name, type_, optional or self.union, nullable, attr, doc)
        self._members.append(member)
        return member

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr()

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, dict):
            value_x = value
        elif mode == ValidationMode.QUERY_STRING and value == '':
            value_x = {}
        else:
            raise _member_error(self, value, _member)

        # Valid union?
        if self.union:
            if len(value_x) != 1:
                raise _member_error(self, value, _member)

        # Result a copy?
        value_copy = None if mode in IMMUTABLE_VALIDATION_MODES else {}

        # Validate all member values
        member_count = 0
        for member in self.members():
            member_name = member.name
            if member_name not in value_x:
                if not member.optional:
                    raise ValidationError(f"Required member {_member_syntax((_member, member_name))!r} missing")
            else:
                member_count += 1
                member_value = value_x[member_name]
                if member.nullable and (member_value is None or \
                        (mode == ValidationMode.QUERY_STRING and not isinstance(member.type, TypeString) and member_value == 'null')):
                    member_value_x = None
                else:
                    member_path = (_member, member_name)
                    member_value_x = member.type.validate(member_value, mode, member_path)
                    if member.attr is not None:
                        member.attr.validate(member_value_x, member_path)
                if value_copy is not None:
                    value_copy[member_name] = member_value_x

        # Any unknown members?
        if member_count != len(value_x):
            member_set = {member.name for member in self.members()}
            unknown_value_names = [value_name for value_name in value_x.keys() if value_name not in member_set]
            raise ValidationError(f"Unknown member {_member_syntax((_member, unknown_value_names[0]))!r:.100s}")

        return value if value_copy is None else value_copy


class TypeArray:
    """
    TODO
    """

    __slots__ = ('type', 'attr')

    #: TODO
    type_name = 'array'

    def __init__(self, type_, attr=None):

        #: TODO
        self.type = type_

        #: TODO
        self.attr = attr

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr(allow_length=True)

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, (list, tuple)):
            value_x = value
        elif mode == ValidationMode.QUERY_STRING and value == '':
            value_x = []
        else:
            raise _member_error(self, value, _member)

        # Result a copy?
        value_copy = None if mode in IMMUTABLE_VALIDATION_MODES else []

        # Validate the list contents
        ix_array_value = 0
        for array_value in value_x:
            member_path = (_member, ix_array_value)
            array_value_x = self.type.validate(array_value, mode, member_path)
            if self.attr is not None:
                self.attr.validate(array_value_x, member_path)
            if value_copy is not None:
                value_copy.append(array_value_x)
            ix_array_value += 1

        return value if value_copy is None else value_copy


class TypeDict:
    """
    TODO
    """

    __slots__ = ('type', 'attr', 'key_type', 'key_attr')

    #: TODO
    type_name = 'dict'

    def __init__(self, type_, attr=None, key_type=None, key_attr=None):

        #: TODO
        self.type = type_

        #: TODO
        self.attr = attr

        #: TODO
        self.key_type = key_type or TYPE_STRING

        #: TODO
        self.key_attr = key_attr

    @staticmethod
    def valid_key_type(key_type):
        """
        TODO
        """

        key_type_base = Typedef.base_type(key_type)
        return isinstance(key_type_base, (TypeString, TypeEnum))

    def has_default_key_type(self):
        """
        TODO
        """

        return isinstance(self.key_type, TypeString)

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr(allow_length=True)

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, dict):
            value_x = value
        elif mode == ValidationMode.QUERY_STRING and value == '':
            value_x = {}
        else:
            raise _member_error(self, value, _member)

        # Result a copy?
        value_copy = None if mode in IMMUTABLE_VALIDATION_MODES else {}

        # Validate the dict key/value pairs
        for dict_key, dict_value in value_x.items():
            member_path = (_member, dict_key)

            # Validate the key
            dict_key_x = self.key_type.validate(dict_key, mode, member_path)
            if self.key_attr is not None:
                self.key_attr.validate(dict_key_x, member_path)

            # Validate the value
            dict_value_x = self.type.validate(dict_value, mode, member_path)
            if self.attr is not None:
                self.attr.validate(dict_value_x, member_path)

            # Result a copy?
            if value_copy is not None:
                value_copy[dict_key_x] = dict_value_x

        return value if value_copy is None else value_copy


class EnumValue:
    """
    An enum value

    :param str value: The enum value string
    :param str doc: The documentation markdown text
    """

    __slots__ = ('value', 'doc')

    def __init__(self, valueString, doc=None):

        #: TODO
        self.value = valueString

        #: The documentation markdown text
        self.doc = doc

    def __eq__(self, other):
        return self.value == other


class TypeEnum:
    """
    An enum type

    :param str type_name: The enum type name
    :param ~chisel.TypeEnum base_types: The enum type's base types
    :param str doc: The documentation markdown text
    """

    __slots__ = ('type_name', 'base_types', '_values', 'doc')

    def __init__(self, type_name='enum', base_types=None, doc=None):

        #: TODO
        self.type_name = type_name

        #: TODO
        self.base_types = base_types

        self._values = []

        #: The documentation markdown text
        self.doc = doc

    def values(self, include_base_types=True):
        """
        TODO
        """

        if include_base_types and self.base_types is not None:
            return chain(
                chain.from_iterable(Typedef.base_type(base_type).values() for base_type in self.base_types if base_type),
                self._values
            )
        return iter(self._values)

    def add_value(self, string, doc=None):
        """
        TODO
        """

        value = EnumValue(string, doc)
        self._values.append(value)
        return value

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr()

    def validate(self, value, unused_mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate the value
        if value not in self.values():
            raise _member_error(self, value, _member)

        return value


class TypeString:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'string'

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr(allow_length=True)

    def validate(self, value, unused_mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate the value
        if not isinstance(value, str):
            raise _member_error(self, value, _member)

        return value


class TypeInt:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'int'

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr(allow_value=True)

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, int) and not isinstance(value, bool):
            value_x = value
        elif isinstance(value, (float, Decimal)):
            value_x = int(value)
            if value_x != value:
                raise _member_error(self, value, _member)
        elif mode == ValidationMode.QUERY_STRING and isinstance(value, str):
            try:
                value_x = int(value)
            except:
                raise _member_error(self, value, _member) from None
        else:
            raise _member_error(self, value, _member)

        return value if mode in IMMUTABLE_VALIDATION_MODES else value_x


class TypeFloat:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'float'

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr(allow_value=True)

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, float):
            value_x = value
        elif isinstance(value, (int, Decimal)) and not isinstance(value, bool):
            value_x = float(value)
        elif mode == ValidationMode.QUERY_STRING and isinstance(value, str):
            try:
                value_x = float(value)
                if isnan(value_x) or isinf(value_x):
                    raise ValueError()
            except:
                raise _member_error(self, value, _member) from None
        else:
            raise _member_error(self, value, _member)

        return value if mode in IMMUTABLE_VALIDATION_MODES else value_x


class TypeBool:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'bool'

    #: TODO
    VALUES = {
        'true': True,
        'false': False
    }

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr()

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, bool):
            return value
        if mode == ValidationMode.QUERY_STRING and isinstance(value, str):
            try:
                return self.VALUES[value]
            except:
                raise _member_error(self, value, _member) from None
        raise _member_error(self, value, _member)


class TypeUuid:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'uuid'

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr()

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, UUID):
            return value
        if mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
            try:
                return UUID(value)
            except:
                raise _member_error(self, value, _member) from None
        raise _member_error(self, value, _member)


class TypeDate:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'date'

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr()

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, date):
            return value
        if mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
            try:
                return date.fromisoformat(value)
            except:
                raise _member_error(self, value, _member) from None
        raise _member_error(self, value, _member)


class TypeDatetime:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'datetime'

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr()

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
        """
        TODO
        """

        # Validate and translate the value
        if isinstance(value, datetime):
            return value
        if mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
            try:
                return datetime.fromisoformat(value)
            except:
                raise _member_error(self, value, _member) from None
        raise _member_error(self, value, _member)


class TypeObject:
    """
    TODO
    """

    __slots__ = ()

    #: TODO
    type_name = 'object'

    @staticmethod
    def validate_attr(attr):
        """
        TODO
        """

        attr.validate_attr()

    def validate(self, value, mode=ValidationMode.DEFAULT, _member=()): # pylint: disable=unused-argument
        """
        TODO
        """

        if value is not None:
            return value
        raise _member_error(self, value, _member)


#: TODO
TYPE_BOOL = TypeBool()

#: TODO
TYPE_DATE = TypeDate()

#: TODO
TYPE_DATETIME = TypeDatetime()

#: TODO
TYPE_FLOAT = TypeFloat()

#: TODO
TYPE_INT = TypeInt()

#: TODO
TYPE_OBJECT = TypeObject()

#: TODO
TYPE_STRING = TypeString()

#: TODO
TYPE_UUID = TypeUuid()



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel request base class and common request classes
"""

from functools import partial
import hashlib
from http import HTTPStatus
import importlib
from itertools import chain
import pkgutil
import posixpath
import re
import sys

from .app import Context


REQUESTS_MODULE_ATTR = '__chisel_requests__'


def request(wsgi_callback=None, **kwargs):
    """
    Decorator for creating a :class:`~chisel.Request` object that wraps a WSGI application function. For example:

    >>> import chisel
    ...
    >>> @chisel.request
    ... def my_request(environ, start_response):
    ...    start_response('200 OK', [('Content-Type', 'text/plain')])
    ...    return [b'Hello, World!']
    ...
    >>> my_request.name, my_request.urls, my_request.doc, my_request.doc_group
    ...
    ('my_request', ((None, '/my_request'),), None, None)

    You can also pass :class:`~chisel.Request` initialization parameters to the request decorator:

    >>> @chisel.request(urls=[('GET', None)], doc='This is my request')
    ... def my_request(environ, start_response):
    ...    start_response('200 OK', [('Content-Type', 'text/plain')])
    ...    return [b'Hello, World!']
    ...
    >>> my_request.name, my_request.urls, my_request.doc, my_request.doc_group
    ...
    ('my_request', (('GET', '/my_request'),), 'This is my request', None)

    The created :class:`~chisel.Request` object is passed to an application's :meth:`~chisel.add_request` method to host
    it with that applicaton.

    >>> application = chisel.Application()
    >>> application.add_request(my_request)
    >>> application.request('GET', '/my_request')
    ('200 OK', [('Content-Type', 'text/plain')], b'Hello, World!')

    :param ~collections.abc.Callable wsgi_callback: A WSGI application function
    :returns Request: The created Request object
    """

    if wsgi_callback is None:
        return partial(request, **kwargs)
    return Request(wsgi_callback, **kwargs).decorate_module(wsgi_callback)


class Request:
    """The Chisel Request object is a wrapper that associates hosting metadata with a WSGI application function.
    See the :func:`~chisel.request` decorator for a common way to create Request objects.

    :param ~collections.abc.Callable wsgi_callback: A WSGI application function
    :param str name: The request name. The default name is the callback function's name.
    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
        or None to match any. The second value is the URL path or None to use the default path.
    :param str doc: The documentation markdown text
    :param str doc_group: The documentation group
    """

    __slots__ = ('wsgi_callback', 'name', 'urls', 'doc', 'doc_group')

    def __init__(self, wsgi_callback=None, name=None, urls=None, doc=None, doc_group=None):
        assert wsgi_callback is not None or name is not None, 'must specify either wsgi_callback and/or name'

        #: The WSGI application function
        self.wsgi_callback = wsgi_callback

        #: The request name
        self.name = name or wsgi_callback.__name__

        #: The documentation markdown text
        self.doc = doc

        #: The documentation group
        self.doc_group = doc_group

        # Normalize urls into list of uppercase-method/path tuple pairs
        if urls is None:
            #: The list of URL method/path tuples
            self.urls = ((None, '/' + self.name),)
        else:
            self.urls = tuple(chain.from_iterable(
                ((None, '/' + self.name),) if url is None else \
                ((url[0] and url[0].upper(), url[1] or '/' + self.name),)
                for url in urls
            ))

    def __call__(self, environ, start_response):
        """
        The request WSGI application method. By default the wrapped wsgi_callback is called. Sub-classes may override this
        method without calling this method.

        :param dict environ: The :pep:`WSGI <3333>` environ dictionary.
        :param ~collections.abc.Callable start_response: The :pep:`WSGI <3333>` start-response callable.
        """

        assert self.wsgi_callback is not None, 'wsgi_callback required when using Request directly'
        return self.wsgi_callback(environ, start_response)

    def decorate_module(self, wsgi_callback):
        """
        Helper method to update a collection requests on the request's module. This information is used by
        :meth:`~chisel.Request.import_requests`.

        :param ~collections.abc.Callable wsgi_callback: The wrapped WSGI application function
        """

        if wsgi_callback.__module__: # pragma: no branch
            module = sys.modules[wsgi_callback.__module__]
            requests = getattr(module, REQUESTS_MODULE_ATTR, None)
            if requests is None:
                requests = {}
                setattr(module, REQUESTS_MODULE_ATTR, requests)
            requests[self.name] = self
        return self

    @staticmethod
    def import_requests(package, parent_package=None):
        """
        Import reqeusts from a package.

        :param str package: The package in which to recursively load requests
        :param str parent_package: The parent package
        """

        package = importlib.import_module(package, parent_package)
        for _, name, _ in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
            module = importlib.import_module(name)
            requests = getattr(module, REQUESTS_MODULE_ATTR, None)
            if requests is not None:
                yield from iter(requests.values())


class RedirectRequest(Request):
    """
    A redirect reqeust

    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
        or None to match any. The second value is the URL path or None to use the default path.
    :param str redirect_url: The redirectd URL
    :param bool permanent: If True, this is a permanent redirect
    :param str name: The request name. By default the name is "redirect_<redirect_url>".
    :param str doc: The documentation markdown text
    :param str doc_group: The documentation group
    """

    __slots__ = ('_status', '_redirect_url')

    def __init__(self, urls, redirect_url, permanent=True, name=None, doc=None, doc_group='Redirects'):
        if name is None:
            name = re.sub(r'([^\w]|_)+', '_', f'redirect_{redirect_url}').rstrip('_')
        if doc is None:
            doc = f'Redirect to {redirect_url}.'
        super().__init__(name=name, urls=urls, doc=doc, doc_group=doc_group)
        self._status = HTTPStatus.MOVED_PERMANENTLY if permanent else HTTPStatus.FOUND
        self._redirect_url = redirect_url

    def __call__(self, environ, unused_start_response):
        ctx = environ[Context.ENVIRON_CTX]
        ctx.add_header('Location', self._redirect_url)
        return ctx.response_text(self._status, self._redirect_url)


class StaticRequest(Request):
    """
    A static resoruce request

    :param str package: The package containing the static resource
    :param str resource_name: The resource name (path)
    :param bool cache: If True, cache the static content. Otherwise, the content is loaded each request.
    :param str content_type: Optional content type string. If None, the content type is auto-determined.
    :param list(tuple) headers: Optional list of key/value header tuples
    :param str name: The request name. By default the name is "static_<redirect_url>".
    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
        or None to match any. The second value is the URL path or None to use the default path.
    :param str doc: The documentation markdown text
    :param str doc_group: The documentation group
    """

    __slots__ = ('_package', '_resource_name', '_cache', '_headers', '_content', '_etag')

    EXT_TO_CONTENT_TYPE = {
        '.css': 'text/css',
        '.html': 'text/html',
        '.jpg': 'image/jpeg',
        '.js': 'application/javascript',
        '.png': 'image/png',
        '.svg': 'image/svg+xml',
        '.txt': 'text/plain',
    }

    def __init__(
            self,
            package,
            resource_name,
            cache=True,
            content_type=None,
            headers=None,
            name=None,
            urls=None,
            doc=None,
            doc_group='Statics'
    ):
        if name is None:
            name = re.sub(r'[^\w]+', '_', resource_name)
        if urls is None:
            urls = [('GET', f'/{resource_name}')]
        if doc is None:
            doc = f'The "{package}" package\'s static resource, "{resource_name}".'
        super().__init__(name=name, urls=urls, doc=doc, doc_group=doc_group)
        self._package = package
        self._resource_name = resource_name
        self._cache = cache
        if content_type is None:
            content_type = self.EXT_TO_CONTENT_TYPE.get(posixpath.splitext(resource_name)[1].lower())
            assert content_type, f'Unknown content type for "{package}" package\'s "{resource_name}" resource'
        self._headers = tuple(chain(headers or (), [('Content-Type', content_type)]))
        if cache:
            self._content, self._etag = self._load_content()

    def _load_content(self):
        content = pkgutil.get_data(self._package, self._resource_name)
        md5 = hashlib.md5()
        md5.update(content)
        return content, md5.hexdigest()

    def __call__(self, environ, start_response):

        # Check the etag - is the resource modified?
        if self._cache:
            content, etag = self._content, self._etag
        else:
            content, etag = self._load_content()
        if etag == environ.get('HTTP_IF_NONE_MATCH'):
            start_response(HTTPStatus.NOT_MODIFIED, [])
            return []

        start_response(HTTPStatus.OK, list(chain(self._headers, [('ETag', etag)])))
        return [content]



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel schema specification language parser
"""

from functools import partial
from itertools import chain
import os
import re

from .model import \
    ActionModel, AttributeValidationError, StructMemberAttributes, \
    TYPE_BOOL, TYPE_DATE, TYPE_DATETIME, TYPE_FLOAT, TYPE_INT, TYPE_OBJECT, TYPE_STRING, TYPE_UUID, \
    TypeArray, TypeDict, TypeEnum, TypeStruct, Typedef


# Spec language regex
RE_PART_ID = r'(?:[A-Za-z]\w*)'
RE_PART_ATTR_GROUP = \
    r'(?:(?P<op><=|<|>|>=|==)\s*(?P<opnum>-?\d+(?:\.\d+)?)' \
    r'|(?P<ltype>len)\s*(?P<lop><=|<|>|>=|==)\s*(?P<lopnum>\d+))'
RE_PART_ATTR = re.sub(r'\(\?P<[^>]+>', r'(?:', RE_PART_ATTR_GROUP)
RE_PART_ATTRS = r'(?:' + RE_PART_ATTR + r'(?:\s*,\s*' + RE_PART_ATTR + r')*)'
RE_ATTR_GROUP = re.compile(RE_PART_ATTR_GROUP)
RE_FIND_ATTRS = re.compile(RE_PART_ATTR + r'(?:\s*,\s*|\s*\Z)')
RE_LINE_CONT = re.compile(r'\\s*$')
RE_COMMENT = re.compile(r'^\s*(?:#-.*|#(?P<doc>.*))?$')
RE_GROUP = re.compile(r'^group(?:\s+"(?P<group>.+?)")?\s*$')
RE_ACTION = re.compile(r'^action\s+(?P<id>' + RE_PART_ID + r')')
RE_PART_BASE_IDS = r'(?:\s*\(\s*(?P<base_ids>' + RE_PART_ID + r'(?:\s*,\s*' + RE_PART_ID + r')*)\s*\)\s*)'
RE_BASE_IDS_SPLIT = re.compile(r'\s*,\s*')
RE_DEFINITION = re.compile(r'^(?P<type>struct|union|enum)\s+(?P<id>' + RE_PART_ID + r')' + RE_PART_BASE_IDS + r'?\s*$')
RE_SECTION = re.compile(r'^\s+(?P<type>path|query|input|output|errors)' + RE_PART_BASE_IDS + r'?\s*$')
RE_SECTION_PLAIN = re.compile(r'^\s+(?P<type>url)\s*$')
RE_PART_TYPEDEF = \
    r'(?P<type>' + RE_PART_ID + r')' \
    r'(?:\s*\(\s*(?P<attrs>' + RE_PART_ATTRS + r')\s*\))?' \
    r'(?:' \
    r'(?:\s*\[\s*(?P<array>' + RE_PART_ATTRS + r'?)\s*\])?' \
    r'|' \
    r'(?:' \
    r'\s*:\s*(?P<dictValueType>' + RE_PART_ID + r')' \
    r'(?:\s*\(\s*(?P<dictValueAttrs>' + RE_PART_ATTRS + r')\s*\))?' \
    r')?' \
    r'(?:\s*\{\s*(?P<dict>' + RE_PART_ATTRS + r'?)\s*\})?' \
    r')' \
    r'\s+(?P<id>' + RE_PART_ID + r')'
RE_TYPEDEF = re.compile(r'^typedef\s+' + RE_PART_TYPEDEF + r'\s*$')
RE_MEMBER = re.compile(r'^\s+(?P<optional>optional\s+)?(?P<nullable>nullable\s+)?' + RE_PART_TYPEDEF + r'\s*$')
RE_VALUE = re.compile(r'^\s+"?(?P<id>(?<!")' + RE_PART_ID + r'(?!")|(?<=").*?(?="))"?\s*$')
RE_URL = re.compile(r'^\s+(?P<method>[A-Za-z]+|\*)(?:\s+(?P<path>/[^\s]*))?')


#: Built-in types
TYPES = {
    'bool': TYPE_BOOL,
    'date': TYPE_DATE,
    'datetime': TYPE_DATETIME,
    'float': TYPE_FLOAT,
    'int': TYPE_INT,
    'object': TYPE_OBJECT,
    'string': TYPE_STRING,
    'uuid': TYPE_UUID
}


class SpecParserError(Exception):
    """
    Chisel specification language parser exception

    :param list(str) errors: The list of error strings
    """

    __slots__ = ('errors',)

    def __init__(self, errors):
        super().__init__('\n'.join(errors))

        #: The list of error strings
        self.errors = errors


class SpecParser:
    """
    The parser class for the Chisel specification language. Parsing can occur at initialization time or by calling the
    :meth:`~chisel.SpecParser.parse_string` method, which can be called repeatedly.

    :param str spec: An optional specification string to parse
    :raises SpecParserError: A parsing error occurred
    """

    __slots__ = (
        'types',
        'actions',
        'errors',
        '_typerefs',
        '_finalize_checks'
    )

    def __init__(self, spec=None):

        #: The dictionary of type name to type object
        self.types = {}

        #: The dictionary of action name to :class:`~chisel.ActionModel`
        self.actions = {}

        #: The list of error strings encountered during parsing
        self.errors = []

        self._typerefs = []
        self._finalize_checks = []

        # Parse the specification string, if any
        if spec is not None:
            self.parse_string(spec)

    def parse(self, lines, filename='', finalize=True):
        """
        Parse a specification from an iterator of line strings (e.g an input stream). This method can be called repeatedly.

        :param ~collections.abc.Iterable(str) lines: An iterator of specification line strings
        :param str filename: The name of file being parsed (for error messages)
        :param bool finalize: If True, resolve names after parsing. Set this argument to False when bulk-parsing related
            specification files. Be sure to call :meth:`~chisel.SpecParser.finalize` when finished.
        :raises SpecParserError: A parsing error occurred
        """

        # Current parser state
        type_ = None
        action = None
        action_sections = None
        urls = None
        doc = []
        doc_group = None
        linenum = 0

        # Helper function get/reset current doc state
        def get_doc():
            nonlocal doc
            if doc:
                doc_str = '\n'.join(doc)
                doc = []
                return doc_str
            return None

        # Process each line
        line_continuation = []
        for line_part in chain(lines, ('',)):
            linenum += 1

            # Line continuation?
            line_part_no_continuation = RE_LINE_CONT.sub('', line_part)
            if line_continuation or line_part_no_continuation is not line_part:
                line_continuation.append(line_part_no_continuation)
            if line_part_no_continuation is not line_part:
                continue
            if line_continuation:
                line = ''.join(line_continuation)
                del line_continuation[:]
            else:
                line = line_part

            # Match syntax
            match_name, match = 'comment', RE_COMMENT.search(line)
            if match is None:
                match_name, match = 'group', RE_GROUP.search(line)
            if match is None:
                match_name, match = 'action', RE_ACTION.search(line)
            if match is None:
                match_name, match = 'definition', RE_DEFINITION.search(line)
            if match is None and action is not None:
                match_name, match = 'section', RE_SECTION.search(line)
            if match is None and action is not None:
                match_name, match = 'section_plain', RE_SECTION_PLAIN.search(line)
            if match is None and isinstance(type_, TypeEnum):
                match_name, match = 'value', RE_VALUE.search(line)
            if match is None and isinstance(type_, TypeStruct):
                match_name, match = 'member', RE_MEMBER.search(line)
            if match is None and urls is not None:
                match_name, match = 'url', RE_URL.search(line)
            if match is None:
                match_name, match = 'typedef', RE_TYPEDEF.search(line)
            if match is None:
                match_name = None

            # Comment?
            if match_name == 'comment':
                doc_string = match.group('doc')
                if doc_string is not None:
                    doc.append(doc_string)

            # Documentation group?
            elif match_name == 'group':
                doc_group = match.group('group')
                if doc_group is not None:
                    doc_group = doc_group.strip()
                else:
                    doc_group = None

            # Action?
            elif match_name == 'action':
                action_id = match.group('id')

                # Action already defined?
                if action_id in self.actions:
                    self._error("Redefinition of action '" + action_id + "'", filename, linenum)

                # Create the new action
                action = ActionModel(action_id, doc=get_doc(), doc_group=doc_group)
                action_sections = set()
                urls = None
                type_ = None
                self.actions[action.name] = action

            # Definition?
            elif match_name == 'definition':
                definition_string = match.group('type')
                definition_id = match.group('id')
                definition_base_ids = match.group('base_ids')
                if definition_base_ids is not None:
                    definition_base_ids = RE_BASE_IDS_SPLIT.split(definition_base_ids)

                # Struct definition
                if definition_string in ('struct', 'union'):

                    # Type already defined?
                    if definition_id in TYPES or definition_id in self.types:
                        self._error("Redefinition of type '" + definition_id + "'", filename, linenum)

                    # Create the new struct type
                    action = None
                    urls = None
                    type_ = TypeStruct(type_name=definition_id, union=(definition_string == 'union'), doc=get_doc())
                    if definition_base_ids is not None:
                        type_.base_types = [None for _ in definition_base_ids]
                        for base_index, base_id in enumerate(definition_base_ids):
                            self._set_type(type_, type_.base_types, base_index, base_id, None, filename, linenum,
                                           self._validate_struct_base_type)
                        self._finalize_checks.append(
                            partial(self._finalize_struct_base_type, None, type_, filename, linenum)
                        )
                    self.types[type_.type_name] = type_

                # Enum definition
                else:  # definition_string == 'enum':

                    # Type already defined?
                    if definition_id in TYPES or definition_id in self.types:
                        self._error("Redefinition of type '" + definition_id + "'", filename, linenum)

                    # Create the new enum type
                    action = None
                    urls = None
                    type_ = TypeEnum(type_name=definition_id, doc=get_doc())
                    if definition_base_ids is not None:
                        type_.base_types = [None for _ in definition_base_ids]
                        for base_index, base_id in enumerate(definition_base_ids):
                            self._set_type(type_, type_.base_types, base_index, base_id, None, filename, linenum,
                                           self._validate_enum_base_type)
                        self._finalize_checks.append(
                            partial(self._finalize_enum_base_type, None, type_, filename, linenum)
                        )
                    self.types[type_.type_name] = type_

            # Section?
            elif match_name == 'section':
                section_string = match.group('type')
                section_base_ids = match.group('base_ids')
                if section_base_ids is not None:
                    section_base_ids = RE_BASE_IDS_SPLIT.split(section_base_ids)

                # Action section redefinition?
                if section_string in action_sections:
                    self._error(f'Redefinition of action {section_string}', filename, linenum)
                action_sections.add(section_string)

                # Set the action section type
                urls = None
                if section_string == 'path':
                    type_ = action.path_type
                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
                elif section_string == 'query':
                    type_ = action.query_type
                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
                elif section_string == 'input':
                    type_ = action.input_type
                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
                elif section_string == 'output':
                    type_ = action.output_type
                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
                else: # section_string == 'errors':
                    type_ = action.error_type
                    validate_base_type_fn, finalize_base_type_fn = self._validate_enum_base_type, self._finalize_enum_base_type

                if section_base_ids is not None:
                    type_.base_types = [None for _ in section_base_ids]
                    for base_index, base_id in enumerate(section_base_ids):
                        self._set_type(
                            type_, type_.base_types, base_index, base_id, None, filename, linenum,
                            partial(validate_base_type_fn, def_type=f'action {section_string}')
                        )
                    self._finalize_checks.append(
                        partial(finalize_base_type_fn, action, type_, filename, linenum)
                    )

            # Plain section?
            elif match_name == 'section_plain':
                section_string = match.group('type')

                # Action section redefinition?
                if section_string in action_sections:
                    self._error(f'Redefinition of action {section_string}', filename, linenum)
                action_sections.add(section_string)

                # Set the action section data
                assert section_string == 'url'
                urls = action.urls
                type_ = None

            # Enum value?
            elif match_name == 'value':
                value_string = match.group('id')

                # Redefinition of enum value?
                if value_string in type_.values(include_base_types=False):
                    self._error("Redefinition of enumeration value '" + value_string + "'", filename, linenum)

                # Add the enum value
                type_.add_value(value_string, doc=get_doc())

            # Struct member?
            elif match_name == 'member':
                optional = match.group('optional') is not None
                nullable = match.group('nullable') is not None
                member_name = match.group('id')

                # Member name already defined?
                if action is None or type_ is action.output_type:
                    struct_members = type_.members
                else:
                    struct_members = action.input_members
                if member_name in (mem.name for mem in struct_members(include_base_types=False)):
                    self._error("Redefinition of member '" + member_name + "'", filename, linenum)

                # Create the member
                member = type_.add_member(member_name, None, optional=optional, nullable=nullable, attr=None, doc=get_doc())
                self._parse_typedef(member, 'type', 'attr', match, filename, linenum)

            # URL?
            elif match_name == 'url':
                method = match.group('method')
                path = match.group('path')
                url = (method if method != '*' else None, path)

                # Duplicate URL?
                if url in urls:
                    self._error(f'Duplicate URL: {method} {"" if path is None else path}', filename, linenum)

                # Add the URL
                urls.append(url)
                get_doc()

            # Typedef?
            elif match_name == 'typedef':
                typedef_name = match.group('id')

                # Type already defined?
                if typedef_name in TYPES or typedef_name in self.types:
                    self._error("Redefinition of type '" + typedef_name + "'", filename, linenum)

                # Create the typedef
                typedef = Typedef(None, attr=None, type_name=typedef_name, doc=get_doc())
                self._parse_typedef(typedef, 'type', 'attr', match, filename, linenum)
                self.types[typedef_name] = typedef

                # Reset current action/type
                action = None
                urls = None
                type_ = None

            # Unrecognized line syntax
            else:
                self._error('Syntax error', filename, linenum)

        # Finalize, if requested
        if finalize:
            self.finalize()

    def parse_string(self, spec, filename='', finalize=True):
        """
        Parse a specification string. This method can be called repeatedly.

        :param str spec: The specification string
        :param str filename: The name of file being parsed (for error messages)
        :param bool finalize: If True, resolve names after parsing. Set this argument to False when bulk-parsing related
            specification files. Be sure to call :meth:`~chisel.SpecParser.finalize` when finished.
        :raises SpecParserError: A parsing error occurred
        """

        self.parse(spec.splitlines(), finalize=finalize, filename=filename)

    def finalize(self):
        """
        Finalize a parsing operation. You only need to call this method if you set :meth:`~chisel.SpecParser.parse`
        "finalize" argument to False.

        :raises SpecParserError: If type name resolution fails
        """

        # Fixup type refs
        for typeref in self._typerefs:
            typeref(True)
        self._typerefs = []

        # Additional finalization checks
        for finalize_check in self._finalize_checks:
            finalize_check()
        self._finalize_checks = []

        # Raise a parser exception if there are any errors
        if self.errors:
            raise SpecParserError(self.errors)

    def load(self, path, spec_ext='.chsl', finalize=True):
        """
        Recursively load and parse specification files. This method can be called repeatedly.

        :param str path: Directory or file path from which to load and parse specification files
        :param str spec_ext: The specification file extension.  The default specification file extension is ".chsl".
        :param bool finalize: If True, resolve names after parsing. Set this argument to False when bulk-parsing related
            specification files. Be sure to call :meth:`~chisel.SpecParser.finalize` when finished.
        :raises SpecParserError: A parsing error occurred
        """

        if os.path.isdir(path):
            for dirpath, _, filenames in os.walk(path):
                for filename in filenames:
                    _, ext = os.path.splitext(filename)
                    if ext == spec_ext:
                        spec_path = os.path.join(dirpath, filename)
                        with open(spec_path, 'r', encoding='utf-8') as file_spec:
                            self.parse(file_spec, filename=spec_path, finalize=False)
        else:
            with open(path, 'r', encoding='utf-8') as file_spec:
                self.parse(file_spec, filename=path, finalize=False)

        if finalize:
            self.finalize()

    # Set a type attribute by name
    def _set_type(self, parent_type, parent_object, parent_type_attr, type_name, type_attr, filename, linenum, validate_fn=None):

        def set_type(error):
            type_ = TYPES.get(type_name) or self.types.get(type_name)
            if type_ is not None:
                error_count = len(self.errors)
                if validate_fn is not None:
                    validate_fn(parent_type, type_, filename, linenum)
                self._validate_attr(type_, type_attr, filename, linenum)
                if error_count == len(self.errors):
                    if isinstance(parent_type_attr, int):
                        parent_object[parent_type_attr] = type_
                    else:
                        setattr(parent_object, parent_type_attr, type_)
            elif error:
                self._error("Unknown member type '" + type_name + "'", filename, linenum)
            return type_

        type_ = set_type(False)
        if type_ is None:
            self._typerefs.append(set_type)

    def _validate_dict_key_type(self, dict_type, key_type, filename, linenum):
        if not dict_type.valid_key_type(key_type):
            self._error('Invalid dictionary key type', filename, linenum)

    def _validate_struct_base_type(self, struct_type, base_type, filename, linenum, def_type=None):
        base_type_base = Typedef.base_type(base_type)
        if not isinstance(base_type_base, TypeStruct) or base_type_base.union != struct_type.union:
            if def_type is None:
                def_type = 'union' if struct_type.union else 'struct'
            self._error(f"Invalid {def_type} base type '{base_type.type_name}'", filename, linenum)

    def _validate_enum_base_type(self, unused_enum_type, base_type, filename, linenum, def_type=None):
        if not isinstance(Typedef.base_type(base_type), TypeEnum):
            if def_type is None:
                def_type = 'enum'
            self._error(f"Invalid {def_type} base type '{base_type.type_name}'", filename, linenum)

    def _finalize_circular_base_type(self, type_, filename, linenum):
        is_circular = False
        base_types = {}
        def traverse_base_types(type_):
            if type_.base_types:
                for base_type in (base_type for base_type in type_.base_types if base_type):
                    base_type_name = Typedef.base_type(base_type).type_name
                    base_type_count = base_types[base_type_name] = base_types.get(base_type_name, 0) + 1
                    if base_type_count == 1:
                        traverse_base_types(Typedef.base_type(base_type))
        traverse_base_types(type_)
        for base_type_name, base_type_count in base_types.items():
            if base_type_count != 1:
                is_circular = True
                self._error("Circular base type detected for type '" + base_type_name + "'", filename, linenum)
        return is_circular

    def _finalize_struct_base_type(self, action, struct_type, filename, linenum):
        if struct_type.base_types is not None and not self._finalize_circular_base_type(struct_type, filename, linenum):
            members = {}
            if action is None or struct_type is action.output_type:
                struct_members = struct_type.members
            else:
                struct_members = action.input_members
            for member in struct_members():
                member_count = members[member.name] = members.get(member.name, 0) + 1
                if member_count == 2:
                    self._error("Redefinition of member '" + member.name + "' from base type", filename, linenum)

    def _finalize_enum_base_type(self, unused_action, enum_type, filename, linenum):
        if enum_type.base_types is not None and not self._finalize_circular_base_type(enum_type, filename, linenum):
            values = {}
            for value in enum_type.values():
                value_count = values[value.value] = values.get(value.value, 0) + 1
                if value_count == 2:
                    self._error("Redefinition of enumeration value '" + value.value + "' from base type", filename, linenum)

    # Record an error
    def _error(self, msg, filename, linenum):
        self.errors.append(f'{filename}:{linenum}: error: {msg}')

    # Validate a type's attributes
    def _validate_attr(self, type_, attr, filename, linenum):
        try:
            if attr is not None:
                type_.validate_attr(attr)
        except AttributeValidationError as exc:
            self._error("Invalid attribute '" + exc.attr + "'", filename, linenum)

    # Parse an attributes string
    @classmethod
    def _parse_attr(cls, attrs_string):
        has_attrs = False
        op_lt, op_lte, op_gt, op_gte, op_eq, op_len_lt, op_len_lte, op_len_gt, op_len_gte, op_len_eq = \
            None, None, None, None, None, None, None, None, None, None
        if attrs_string is not None:
            for attr_string in RE_FIND_ATTRS.findall(attrs_string):
                has_attrs = True
                match_attr = RE_ATTR_GROUP.match(attr_string)
                attr_op = match_attr.group('op')
                attr_length_op = match_attr.group('lop') if attr_op is None else None

                if attr_op is not None:
                    attr_value = float(match_attr.group('opnum'))
                    if attr_op == '<':
                        op_lt = attr_value
                    elif attr_op == '<=':
                        op_lte = attr_value
                    elif attr_op == '>':
                        op_gt = attr_value
                    elif attr_op == '>=':
                        op_gte = attr_value
                    else:  # ==
                        op_eq = attr_value
                else:  # attr_length_op is not None:
                    attr_value = int(match_attr.group('lopnum'))
                    if attr_length_op == '<':
                        op_len_lt = attr_value
                    elif attr_length_op == '<=':
                        op_len_lte = attr_value
                    elif attr_length_op == '>':
                        op_len_gt = attr_value
                    elif attr_length_op == '>=':
                        op_len_gte = attr_value
                    else:  # ==
                        op_len_eq = attr_value

        if not has_attrs:
            return None
        return StructMemberAttributes(op_eq, op_lt, op_lte, op_gt, op_gte, op_len_eq, op_len_lt, op_len_lte, op_len_gt, op_len_gte)

    # Construct typedef parts
    def _parse_typedef(self, parent, parent_type_attr, parent_attr_attr, match_typedef, filename, linenum):
        array_attrs_string = match_typedef.group('array')
        dict_attrs_string = match_typedef.group('dict')

        # Array member?
        if array_attrs_string is not None:
            value_type_name = match_typedef.group('type')
            value_attr = self._parse_attr(match_typedef.group('attrs'))
            array_type = TypeArray(None, attr=value_attr)
            self._set_type(array_type, array_type, 'type', value_type_name, value_attr, filename, linenum)

            array_attr = self._parse_attr(array_attrs_string)
            self._validate_attr(array_type, array_attr, filename, linenum)

            setattr(parent, parent_type_attr, array_type)
            setattr(parent, parent_attr_attr, array_attr)

        # Dictionary member?
        elif dict_attrs_string is not None:
            value_type_name = match_typedef.group('dictValueType')
            if value_type_name is not None:
                value_attr = self._parse_attr(match_typedef.group('dictValueAttrs'))
                key_type_name = match_typedef.group('type')
                key_attr = self._parse_attr(match_typedef.group('attrs'))
                dict_type = TypeDict(None, attr=value_attr, key_type=None, key_attr=key_attr)
                self._set_type(dict_type, dict_type, 'type', value_type_name, value_attr, filename, linenum)
                self._set_type(dict_type, dict_type, 'key_type', key_type_name, key_attr, filename, linenum, self._validate_dict_key_type)
            else:
                value_type_name = match_typedef.group('type')
                value_attr = self._parse_attr(match_typedef.group('attrs'))
                dict_type = TypeDict(None, attr=value_attr)
                self._set_type(dict_type, dict_type, 'type', value_type_name, value_attr, filename, linenum)

            dict_attr = self._parse_attr(dict_attrs_string)
            self._validate_attr(dict_type, dict_attr, filename, linenum)

            setattr(parent, parent_type_attr, dict_type)
            setattr(parent, parent_attr_attr, dict_attr)

        # Non-container member...
        else:
            member_type_name = match_typedef.group('type')
            member_attr = self._parse_attr(match_typedef.group('attrs'))

            self._set_type(parent, parent, parent_type_attr, member_type_name, member_attr, filename, linenum)
            setattr(parent, parent_attr_attr, member_attr)



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel is a light-weight Python WSGI application framework with tools for building well-documented,
well-tested, schema-validated JSON web APIs.
"""

__version__ = '0.9.98'

from .action import \
    Action, \
    ActionError, \
    action

from .app import \
    Application, \
    Context

from .doc import \
    create_doc_requests

from .model import \
    ActionModel, \
    AttributeValidationError, \
    EnumValue, \
    StructMember, \
    StructMemberAttributes, \
    TYPE_BOOL, \
    TYPE_DATE, \
    TYPE_DATETIME, \
    TYPE_FLOAT, \
    TYPE_INT, \
    TYPE_OBJECT, \
    TYPE_STRING, \
    TYPE_UUID, \
    TypeArray, \
    TypeBool, \
    TypeDate, \
    TypeDatetime, \
    TypeDict, \
    TypeEnum, \
    TypeFloat, \
    TypeInt, \
    TypeObject, \
    TypeString, \
    TypeStruct, \
    TypeUuid, \
    Typedef, \
    ValidationError, \
    ValidationMode, \
    get_referenced_types

from .request import \
    RedirectRequest, \
    Request, \
    StaticRequest, \
    request

from .spec import \
    SpecParser, \
    SpecParserError

from .util import \
    JSONEncoder, \
    decode_query_string, \
    encode_query_string



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel documentation application
"""

from collections import defaultdict

from .action import Action, ActionError
from .model import Typedef, TypeStruct, TypeEnum, TypeArray, TypeDict, get_referenced_types
from .request import RedirectRequest, StaticRequest


def create_doc_requests(requests=None, root_path='/doc', api=True, app=True, css=True, cache=True):
    """
    Yield a series of requests for use with :meth:`~chisel.Application.add_requests` comprising the Chisel
    documentation application. By default, the documenation application is hosted at "/doc/".

    :param requests: A list of requests or None to use the application's requests
    :type requests: list(~chisel.Request)
    :param str root_path: The documentation application URL root path. The default is "/doc".
    :param bool api: If True, include the documentation APIs. Two documentation APIs are added,
        "/doc/doc_index" and "`/doc/doc_request <doc/doc.html#name=chisel_doc_request>`_".
    :param bool app: If True, include the documentation client application.
    :param bool css: If True, and "app" is True, include "/doc/doc.css" and "/doc/doc.svg". If you exclude CSS,
        you can add your own versions of these requests to customize the documentation styling.
    :param bool cache: If True, cache static request content.
    :returns: Generator of :class:`~chisel.Request`
    """

    if api:
        yield DocIndex(requests=requests, urls=(('GET', root_path + '/doc_index'),))
        yield DocRequest(requests=requests, urls=(('GET', root_path + '/doc_request/{name}'),))
    if app:
        yield RedirectRequest((('GET', root_path),), root_path + '/')
        yield StaticRequest('chisel', 'static/doc.html', urls=(('GET', root_path + '/'), ('GET', root_path + '/index.html')), cache=cache)
        if css:
            yield StaticRequest('chisel', 'static/doc.css', urls=(('GET', root_path + '/doc.css'),), cache=cache)
            yield StaticRequest('chisel', 'static/doc.svg', urls=(('GET', root_path + '/doc.svg'),), cache=cache)
        yield StaticRequest('chisel', 'static/chisel.js', urls=(('GET', root_path + '/chisel.js'),), cache=cache)
        yield StaticRequest('chisel', 'static/doc.js', urls=(('GET', root_path + '/doc.js'),), cache=cache)


class DocIndex(Action):
    """
    The documentation index API. This API provides all the information the documentation application needs to render the
    index page.

    :param requests: A list of requests or None to use the application's requests
    :type requests: dict(str, ~chisel.Request)
    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
        or None to match any. The second value is the URL path or None to use the default path.
    """

    __slots__ = ('requests',)

    SPEC = '''\
group "Documentation"

# A non-empty string array
typedef string[len > 0] StringArray

# Get the documentation index
action chisel_doc_index

    output

        # The documentation index title
        string title

        # The dictionary of documentation group titles to array of request names
        StringArray{} groups
'''

    def __init__(self, requests=None, urls=(('GET', '/doc_index'),)):
        super().__init__(self._doc_index, name='chisel_doc_index', urls=urls, spec=self.SPEC)
        if requests is not None:
            self.requests = {request.name: request for request in requests}
        else:
            self.requests = None

    def _doc_index(self, ctx, unused_req):
        requests = self.requests if self.requests is not None else ctx.app.requests
        groups = defaultdict(list)
        for request in requests.values():
            groups[request.doc_group or 'Uncategorized'].append(request.name)
        return {
            'title': ctx.environ['HTTP_HOST'],
            'groups': {group: sorted(names) for group, names in groups.items()}
        }


class DocRequest(Action):
    """
    The documentation request API. This API provides all the information the documentation applicaton needs to render
    the request documentation page. The documentation request API's documentation is `here
    <doc/doc.html#name=chisel_doc_request>`_.

    :param requests: A list of requests or None to use the application's requests
    :type requests: list(~chisel.Request)
    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
        or None to match any. The second value is the URL path or None to use the default path.
    """

    __slots__ = ('requests',)

    SPEC = '''
group "Documentation"

# Union representing a member type
union Type

    # A built-in type
    BuiltinType builtin

    # An array type
    Array array

    # A dictionary type
    Dict dict

    # An enumeration type
    string enum

    # A struct type
    string struct

    # A type definition
    string typedef

# A type attribute
struct Attribute

    # The value is equal
    optional float eq

    # The value is less than
    optional float lt

    # The value is less than or equal to
    optional float lte

    # The value is greater than
    optional float gt

    # The value is greater than or equal to
    optional float gte

    # The length is equal to
    optional int len_eq

    # The length is less-than
    optional int len_lt

    # The length is less than or equal to
    optional int len_lte

    # The length is greater than
    optional int len_gt

    # The length is greater than or equal to
    optional int len_gte

# The built-in type enumeration
enum BuiltinType

    # The string type
    string

    # The integer type
    int

    # The float type
    float

    # The boolean type
    bool

    # A date formatted as an ISO-8601 date string
    date

    # A date/time formatted as an ISO-8601 date/time string
    datetime

    # A UUID formatted as string
    uuid

    # An object of any type
    object

# An array type
struct Array

    # The contained type
    Type type

    # The contained type's attributes
    optional Attribute attr

# A dictionary type
struct Dict

    # The contained key type
    Type type

    # The contained key type's attributes
    optional Attribute attr

    # The contained value type
    Type key_type

    # The contained value type's attributes
    optional Attribute key_attr

# An enumeration type
struct Enum

    # The documentation markdown text
    optional string doc

    # The enum type name
    string name

    # The enumeration values
    EnumValue[] values

# An enumeration type value
struct EnumValue

    # The documentation markdown text
    optional string doc

    # The value string
    string value

# A struct type
struct Struct

    # The documentation markdown text
    optional string doc

    # The struct type name
    string name

    # The struct members
    Member[] members

    # If true, the struct is a union and exactly one of the optional members is present.
    optional bool union

# A struct type member
struct Member

    # The documentation markdown text
    optional string doc

    # The member name
    string name

    # The member type
    Type type

    # The member type attributes
    optional Attribute attr

    # If true, the member is optional and may not be present.
    optional bool optional

    # If true, the member may be null.
    optional bool nullable

# A typedef type
struct Typedef

    # The documentation markdown text
    optional string doc

    # The typedef type name
    string name

    # The typedef's type
    Type type

    # The typedef's type attributes
    optional Attribute attr

# A JSON web service API
struct Action

    # The action name
    string name

    # The path parameters struct
    Struct path

    # The query parameters struct
    Struct query

    # The content body struct
    Struct input

    # The response body struct
    optional Struct output

    # The custom error response codes
    Enum errors

# Struct representing a request's URL information
struct RequestURL

    # The request URL HTTP request method. If not present, all HTTP request methods are accepted.
    optional string method

    # The request URL path
    string url

# Get a request's documentation information
action chisel_doc_request

    path

        # The request name
        string name

    output

        # The documentation markdown text
        optional string doc

        # The request name
        string name

        # The array of URL paths where the request is hosted.
        RequestURL[] urls

        # The action definition. If this member is present then this request is an action.
        optional Action action

        # The array of struct types referenced by the action
        optional Struct[] structs

        # The array of enum types referenced by the action
        optional Enum[] enums

        # The array of typedef types referenced by the action
        optional Typedef[] typedefs

    errors

        # The request name is unknown
        UnknownName
'''

    def __init__(self, requests=None, urls=(('GET', '/doc_request'),)):
        super().__init__(self._doc_request, name='chisel_doc_request', urls=urls, spec=self.SPEC)
        if requests is not None:
            #: Optional list of requests to document or None. If None, the applications request collection is used.
            self.requests = {request.name: request for request in requests}
        else:
            self.requests = None

    def _doc_request(self, ctx, req):
        requests = self.requests if self.requests is not None else ctx.app.requests
        request = requests.get(req['name'])
        if request is None:
            raise ActionError('UnknownName')

        response = {
            'name': request.name,
            'urls': [self._url_dict(method, url) for method, url in request.urls],
        }
        if request.doc is not None:
            response['doc'] = request.doc

        if isinstance(request, Action):
            response['action'] = action_dict = {
                'name': request.model.name,
                'path': self._struct_dict(request.model.path_type),
                'query': self._struct_dict(request.model.query_type),
                'input': self._struct_dict(request.model.input_type),
                'errors': self._enum_dict(request.model.error_type),
            }
            if not request.wsgi_response:
                action_dict['output'] = self._struct_dict(request.model.output_type)

            response_structs = sorted(
                (self._struct_dict(typ) for typ in get_referenced_types(request.model) if isinstance(typ, TypeStruct)),
                key=lambda x: x['name']
            )
            if response_structs:
                response['structs'] = response_structs

            response_enums = sorted(
                (self._enum_dict(typ) for typ in get_referenced_types(request.model) if isinstance(typ, TypeEnum)),
                key=lambda x: x['name']
            )
            if response_enums:
                response['enums'] = response_enums

            response_typedefs = sorted(
                (self._typedef_dict(typ) for typ in get_referenced_types(request.model) if isinstance(typ, Typedef)),
                key=lambda x: x['name']
            )
            if response_typedefs:
                response['typedefs'] = response_typedefs

        return response

    @staticmethod
    def _url_dict(method, url):
        url_dict = {'url': url}
        if method is not None:
            url_dict['method'] = method
        return url_dict

    @classmethod
    def _type_dict(cls, type_):
        if isinstance(type_, TypeArray):
            return {'array': cls._array_dict(type_)}
        if isinstance(type_, TypeDict):
            return {'dict': cls._dict_dict(type_)}
        if isinstance(type_, TypeEnum):
            return {'enum': type_.type_name}
        if isinstance(type_, TypeStruct):
            return {'struct': type_.type_name}
        if isinstance(type_, Typedef):
            return {'typedef': type_.type_name}
        return {'builtin': type_.type_name}

    @classmethod
    def _array_dict(cls, array_type):
        array_dict = {
            'type': cls._type_dict(array_type.type),
        }
        if array_type.attr is not None:
            array_dict['attr'] = cls._attr_dict(array_type.attr)
        return array_dict

    @classmethod
    def _dict_dict(cls, dict_type):
        dict_dict = {
            'type': cls._type_dict(dict_type.type),
            'key_type': cls._type_dict(dict_type.key_type),
        }
        if dict_type.attr is not None:
            dict_dict['attr'] = cls._attr_dict(dict_type.attr)
        if dict_type.key_attr is not None:
            dict_dict['key_attr'] = cls._attr_dict(dict_type.key_attr)
        return dict_dict

    @classmethod
    def _attr_dict(cls, attr):
        attr_dict = {}
        if attr.op_eq is not None:
            attr_dict['eq'] = attr.op_eq
        if attr.op_lt is not None:
            attr_dict['lt'] = attr.op_lt
        if attr.op_lte is not None:
            attr_dict['lte'] = attr.op_lte
        if attr.op_gt is not None:
            attr_dict['gt'] = attr.op_gt
        if attr.op_gte is not None:
            attr_dict['gte'] = attr.op_gte
        if attr.op_len_eq is not None:
            attr_dict['len_eq'] = attr.op_len_eq
        if attr.op_len_lt is not None:
            attr_dict['len_lt'] = attr.op_len_lt
        if attr.op_len_lte is not None:
            attr_dict['len_lte'] = attr.op_len_lte
        if attr.op_len_gt is not None:
            attr_dict['len_gt'] = attr.op_len_gt
        if attr.op_len_gte is not None:
            attr_dict['len_gte'] = attr.op_len_gte
        return attr_dict

    @classmethod
    def _struct_dict(cls, struct_type):
        struct_dict = {
            'name': struct_type.type_name,
            'members': [cls._member_dict(member) for member in struct_type.members()],
        }
        if struct_type.doc is not None:
            struct_dict['doc'] = struct_type.doc
        if struct_type.union:
            struct_dict['union'] = True
        return struct_dict

    @classmethod
    def _member_dict(cls, member):
        member_dict = {
            'name': member.name,
            'type': cls._type_dict(member.type),
        }
        if member.doc is not None:
            member_dict['doc'] = member.doc
        if member.optional:
            member_dict['optional'] = member.optional
        if member.nullable:
            member_dict['nullable'] = member.nullable
        if member.attr is not None:
            member_dict['attr'] = cls._attr_dict(member.attr)
        return member_dict

    @classmethod
    def _enum_dict(cls, enum_type):
        enum_dict = {
            'name': enum_type.type_name,
            'values': [cls._enum_value_dict(enum_value) for enum_value in enum_type.values()],
        }
        if enum_type.doc is not None:
            enum_dict['doc'] = enum_type.doc
        return enum_dict

    @classmethod
    def _enum_value_dict(cls, enum_value):
        enum_value_dict = {
            'value': enum_value.value,
        }
        if enum_value.doc is not None:
            enum_value_dict['doc'] = enum_value.doc
        return enum_value_dict

    @classmethod
    def _typedef_dict(cls, typedef_type):
        typedef_dict = {
            'name': typedef_type.type_name,
            'type': cls._type_dict(typedef_type.type),
        }
        if typedef_type.doc is not None:
            typedef_dict['doc'] = typedef_type.doc
        if typedef_type.attr is not None:
            typedef_dict['attr'] = cls._attr_dict(typedef_type.attr)
        return typedef_dict



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel WSGI application base class and utilities
"""

from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from io import BytesIO
import logging
import re
from urllib.parse import quote, unquote

from .util import encode_query_string, JSONEncoder


# Regular expression for matching URL arguments
RE_URL_ARG = re.compile(r'/\{([A-Za-z]\w*)\}')
RE_URL_ARG_ESC = re.compile(r'/\\{([A-Za-z]\w*)\\}')


class Application:
    """
    The chisel application base class. Override this class if you need additional state for your application (like
    application configuration). Add functionality to your application by adding :class:`~chisel.Request` or
    :class:`~chisel.Action` objects to it using :func:`~chisel.Application.add_request` or
    :func:`~chisel.Application.add_requests`).
    """

    __slots__ = (
        'log_level',
        'log_format',
        'pretty_output',
        'validate_output',
        'requests',
        '__request_urls',
        '__request_regex'
    )

    def __init__(self):

        #: The application's `logging level <https://docs.python.org/3/library/logging.html#logging-levels>`_.
        #: The default is :py:data:`logging.WARNING`.
        self.log_level = logging.WARNING

        #: The application's log format string. The default is ``'%(levelname)s [%(process)s / %(thread)s] %(message)s'``.
        self.log_format = '%(levelname)s [%(process)s / %(thread)s] %(message)s'

        #: Set to True for "pretty" request output. Individual requests can : use this application state they see
        #: fit. For example, :class:`~chisel.Action` : requests return indented JSON when this value is True. Default is
        #: False.
        self.pretty_output = False

        #: Set to True to validate request output. Individual requests can use this application state as they see fit. For
        #: example, :class:`~chisel.Action` requests schema-validate return content data when this value is True.
        #: is True.
        self.validate_output = True

        #: The chisel application's map of request name to :class:`~chisel.Request` object map.
        self.requests = {}

        self.__request_urls = {}
        self.__request_regex = []

    def add_request(self, request):
        """
        Add a :class:`~chisel.Request` to the application.

        :param ~chisel.Request request: The request object.
        """

        # Duplicate request name?
        if request.name in self.requests:
            raise ValueError(f'redefinition of request "{request.name}"')
        self.requests[request.name] = request

        # Add the request URLs
        for method, url in request.urls:

            # URL with arguments?
            if RE_URL_ARG.search(url):
                request_regex = '^' + RE_URL_ARG_ESC.sub(r'/(?P<\1>[^/]+)', re.escape(url)) + '$'
                self.__request_regex.append((method, re.compile(request_regex), request))
            else:
                request_key = (method, url)
                if request_key in self.__request_urls:
                    raise ValueError(f'redefinition of request URL "{url}"')
                self.__request_urls[request_key] = request

    def add_requests(self, requests):
        """
        Add a series of :class:`~chisel.Request` objects to the application.

        :param ~collections.abc.Iterable(~chisel.Request) requests: A list of :class:`~chisel.Request` objects.
        """

        for request in requests:
            self.add_request(request)

    def __call__(self, environ, start_response):
        """
        The chisel application WSGI callback. When the application recieves an HTTP request, this method matches the
        appropriate :class:`~chisel.Request' object and then calls its :func:`~chisel.Request.__call__` method. The
        application and URL path arguments (e.g. ``'/documents/{id}'``) are made available to the request through the
        request's :class:`~chisel.Context` object.

        :param dict environ: The :pep:`WSGI <3333>` environ dictionary
        :param ~collections.abc.Callable start_response: The :pep:`WSGI <3333>` start-response callable
        :returns: The WSGI content iterable
        """

        # Match the request by method and exact URL
        path_info = environ['PATH_INFO']
        request_method = environ['REQUEST_METHOD'].upper()
        is_head = (request_method == 'HEAD')
        if is_head:
            request_method = environ['REQUEST_METHOD'] = 'GET'
        request, url_args = self.__request_urls.get((request_method, path_info)), None
        if request is None:

            # Match the request by method and URL regex
            request, url_args = next(
                (
                    (request, {unquote(url_arg): unquote(url_value) for url_arg, url_value in request_match.groupdict().items()})
                    for request, request_match in
                    (
                        (request, regex.match(path_info)) for method, regex, request in self.__request_regex
                        if method is not None and method == request_method
                    )
                    if request_match
                ),
                (None, None)
            )
            if request is None:

                # Match the request by exact URL (any method)
                request, url_args = self.__request_urls.get((None, path_info)), None
                if request is None:

                    # Match the request by URL regex (any method)
                    request, url_args = next(
                        (
                            (request, {unquote(url_arg): unquote(url_value) for url_arg, url_value in request_match.groupdict().items()})
                            for request, request_match in
                            (
                                (request, regex.match(path_info)) for method, regex, request in self.__request_regex
                                if method is None
                            )
                            if request_match
                        ),
                        (None, None)
                    )

        # Create the request context
        ctx = environ[Context.ENVIRON_CTX] = Context(self, environ, start_response, url_args)

        # Request not found?
        if request is None:
            if any(path == path_info for _, path in self.__request_urls) or \
               any(regex.match(path_info) for _, regex, _ in self.__request_regex):
                response = ctx.response_text(HTTPStatus.METHOD_NOT_ALLOWED)
            else:
                response = ctx.response_text(HTTPStatus.NOT_FOUND)
        else:
            # Handle the request
            try:
                response = request(ctx.environ, ctx.start_response)
            except: # pylint: disable=bare-except
                ctx.log.exception('exception raised by request "%s"', request.name)
                response = ctx.response_text(HTTPStatus.INTERNAL_SERVER_ERROR)

        if is_head:
            return []
        return response

    def request(self, request_method, path_info, query_string='', wsgi_input=b'', environ=None):
        """
        Execute an application request

        :param str request_method: The HTTP request method string (e.g. ``'GET'``)
        :param str path_info: The request URL path (e.g. ``'/doc/'``)
        :param str query_string: Optional query string
        :param bytes wsgi_input: Optional request content
        :param dict environ: Optional environ dict. If not provided, a minimal default environ is created.
        :returns: Response status, headers, and content bytes
        """

        request_environ = Context.create_environ(request_method, path_info, query_string, wsgi_input, environ=environ)
        start_response = StartResponse()
        response = self(request_environ, start_response)
        return start_response.status, start_response.headers, b''.join(response)


class Context:
    """
    Class to encapsulate HTTP request state. :class:`~chisel.Application` passes a Context object to each request in
    the WSGI environ dict, ``environ[chisel.Context.ENVIRON_CTX]``.

    :param ~chisel.Application app: The chisel application object
    :param dict environ: The :pep:`WSGI <3333>` environ dictionary
    :param ~collections.abc.Callable start_response: The :pep:`WSGI <3333>` start-response callable
    :param dict url_args: The parsed URL arguments dictionary
    """

    __slots__ = ('app', 'environ', '_start_response', 'url_args', 'log', 'headers')

    #: The context WSGI environ key
    ENVIRON_CTX = 'chisel.ctx'

    def __init__(self, app, environ=None, start_response=None, url_args=None):

        #: The :class:`~chisel.Application` serving the request
        self.app = app

        #: The WSGI environ dictionary
        self.environ = environ or {}

        self._start_response = start_response

        #: The URL path arguments, if any
        self.url_args = url_args

        #: The request's header map. These headers are added to the response.
        self.headers = {}

        #: The python logger instance. Write log messages using this object directly.
        self.log = logging.getLoggerClass()('')
        self.log.setLevel(app.log_level)
        wsgi_errors = environ.get('wsgi.errors') if environ else None
        if wsgi_errors is None:
            handler = logging.NullHandler()
        else:
            handler = logging.StreamHandler(wsgi_errors)
        if callable(app.log_format):
            handler.setFormatter(app.log_format(self))
        else:
            handler.setFormatter(logging.Formatter(app.log_format))
        self.log.addHandler(handler)

    @staticmethod
    def create_environ(request_method, path_info, query_string='', wsgi_input=b'', environ=None):
        """
        Create a minimal, test WSGI environ dict

        :param str request_method: The HTTP request method (e.g. ``'GET'``)
        :param str path_info: The request URL path (e.g. ``'/doc/'``)
        :param str query_string: Optional query string
        :param bytes wsgi_input: Optional request content
        :param dict environ: Optional environ dict. If not provided, a minimal default environ is created.
        :returns: The created :class:`~chisel.Context` object
        """

        if environ is None:
            environ = {}
        environ.setdefault('HTTP_HOST', 'localhost:80')
        environ.setdefault('PATH_INFO', path_info)
        environ.setdefault('QUERY_STRING', query_string if isinstance(query_string, str) else encode_query_string(query_string))
        environ.setdefault('REQUEST_METHOD', request_method)
        environ.setdefault('SCRIPT_NAME', '')
        environ.setdefault('SERVER_NAME', 'localhost')
        environ.setdefault('SERVER_PORT', '80')
        environ.setdefault('wsgi.input', BytesIO(wsgi_input))
        environ.setdefault('wsgi.url_scheme', 'http')
        return environ

    def start_response(self, status, headers):
        """
        Call start response on the WSGI request's start_response function. Any headers added using
        :meth:`~chisel.Context.add_header` are included.

        :param status: The response HTTP status (e.g. "HTTPStatus.OK")
        :type status: ~http.HTTPStatus or str
        :param list(tuple) headers: List of key/value header tuples
        """

        if not isinstance(status, str):
            status = f'{status.value} {status.phrase}'
        for key, value in headers:
            self.add_header(key, value)
        self._start_response(status, sorted(self.headers.items()))

    def add_header(self, key, value):
        """
        Add a header key/value to the request's response

        >>> import chisel
        ...
        >>> @chisel.action(spec='''
        ... action my_action
        ...     url
        ...         GET
        ... ''')
        ... def my_action(ctx, req):
        ...    ctx.add_header('ETag', 'foo')
        ...    return {}
        ...
        >>> application = chisel.Application()
        >>> application.add_request(my_action)
        >>> application.request('GET', '/my_action')
        ('200 OK', [('Content-Type', 'application/json'), ('ETag', 'foo')], b'{}')

        :param str key: The header key
        :param str value: The header value
        """

        assert isinstance(key, str), 'header key must be of type str'
        assert isinstance(value, str), 'header value must be of type str'
        self.headers[key] = value

    def add_cache_headers(self, control, ttl_seconds=None, utcnow=None):
        """
        Add a cache header to the response. You can specify a public or private cache with a time-to-live. You can specify
        no-cache by passing control as None.

        >>> from datetime import datetime
        >>> from pprint import pprint
        >>> import chisel
        ...
        >>> @chisel.action(spec='''
        ... action my_action
        ...     url
        ...         GET
        ... ''')
        ... def my_action(ctx, req):
        ...    ctx.add_cache_headers('private', ttl_seconds=300, utcnow=datetime.fromisoformat('2020-05-19T17:19:00-07:00'))
        ...    return {}
        ...
        >>> application = chisel.Application()
        >>> application.add_request(my_action)
        >>> pprint(application.request('GET', '/my_action'))
        ('200 OK',
         [('Cache-Control', 'private,max-age=300'),
          ('Content-Type', 'application/json'),
          ('Expires', 'Wed, 20 May 2020 00:24:00 GMT')],
         b'{}')

        :param str control: ``'public'``, ``'private'``, or None (for no-cache)
        :param int ttl_seconds: Cache duration in seconds. Do not specify for no-cache.
        :param ~datetime.datetime utcnow: A :func:`~datetime.datetime` to use as the current datetime
        """

        if self.environ.get('REQUEST_METHOD') == 'GET':
            if control is None:
                self.add_header('Cache-Control', 'no-cache')
            else:
                assert control in ('public', 'private')
                assert isinstance(ttl_seconds, int) and ttl_seconds > 0
                self.add_header('Cache-Control', f'{control},max-age={ttl_seconds}')
                if utcnow is None:
                    utcnow = datetime.utcnow()
                else:
                    utcnow = utcnow.astimezone(timezone.utc)
                self.add_header('Expires', (utcnow + timedelta(seconds=ttl_seconds)).strftime('%a, %d %b %Y %H:%M:%S GMT'))

    def response(self, status, content_type, content, headers=None):
        """
        A generic WSGI response

        >>> from http import HTTPStatus
        >>> import chisel
        ...
        >>> @chisel.action(wsgi_response=True, spec='''
        ... action my_action
        ...     url
        ...         GET
        ... ''')
        ... def my_action(ctx, req):
        ...    return ctx.response(HTTPStatus.OK, 'text/plain', [b'Hello'])
        ...
        >>> application = chisel.Application()
        >>> application.add_request(my_action)
        >>> application.request('GET', '/my_action')
        ('200 OK', [('Content-Type', 'text/plain')], b'Hello')

        :param status: The HTTP response status
        :type status: ~http.HTTPStatus or str
        :param str content_type: The response content type
        :param ~collections.abc.Iterable(bytes) content: The WSGI response content
        :param list(tuple) headers: Optional list of key/value header tuples to add to the response
        :returns: The WSGI response content iterable
        """

        assert not isinstance(content, (str, bytes)), 'response content cannot be of type str or bytes'
        response_headers = [('Content-Type', content_type)]
        if headers:
            response_headers.extend(headers)
        self.start_response(status, response_headers)
        return content

    def response_text(self, status, text=None, content_type='text/plain', encoding='utf-8', headers=None):
        """
        A plain-text WSGI response

        >>> from http import HTTPStatus
        >>> import chisel
        ...
        >>> @chisel.action(wsgi_response=True, spec='''
        ... action my_action
        ...     url
        ...         GET
        ... ''')
        ... def my_action(ctx, req):
        ...    return ctx.response_text(HTTPStatus.OK, "Hello")
        ...
        >>> application = chisel.Application()
        >>> application.add_request(my_action)
        >>> application.request('GET', '/my_action')
        ('200 OK', [('Content-Type', 'text/plain')], b'Hello')

        :param status: The HTTP response status
        :type status: ~http.HTTPStatus or str
        :param str text: The response text
        :param str content_type: The response content type. The default is "text/plain".
        :param str encoding: The content encoding. The default is "utf-8".
        :param list(tuple) headers: Optional list of key/value header tuples to add to the response
        """

        if text is None:
            if isinstance(status, str):
                text = status
            else:
                text = status.phrase
        return self.response(status, content_type, [text.encode(encoding)], headers=headers)

    def response_json(self, status, response, content_type='application/json', encoding='utf-8', headers=None, jsonp=None):
        """
        A JSON response

        >>> from http import HTTPStatus
        >>> import chisel
        ...
        >>> @chisel.action(wsgi_response=True, spec='''
        ... action my_action
        ...     url
        ...         GET
        ... ''')
        ... def my_action(ctx, req):
        ...    return ctx.response_json(HTTPStatus.OK, {'a': 5, 'b': 7})
        ...
        >>> application = chisel.Application()
        >>> application.add_request(my_action)
        >>> application.request('GET', '/my_action')
        ('200 OK', [('Content-Type', 'application/json')], b'{"a":5,"b":7}')

        :param status: The HTTP response status
        :type status: ~http.HTTPStatus or str
        :param dict response: The response dictionary
        :param str content_type: The response content type. The default is "application/json".
        :param str encoding: The content encoding. The default is "utf-8".
        :param list(tuple) headers: Optional list of key/value header tuples to add to the response
        :param str jsonp: Optional JSONP key
        """

        encoder = JSONEncoder(
            check_circular=self.app.validate_output,
            allow_nan=False,
            sort_keys=True,
            indent=2 if self.app.pretty_output else None,
            separators=(',', ': ') if self.app.pretty_output else (',', ':')
        )
        content = encoder.encode(response)
        if jsonp:
            content_list = [jsonp.encode(encoding), b'(', content.encode(encoding), b');']
        else:
            content_list = [content.encode(encoding)]
        return self.response(status, content_type, content_list, headers=headers)

    def reconstruct_url(self, path_info=None, query_string=None, relative=False):
        """
        Reconstruct the request's URL

        >>> import chisel
        >>> application = chisel.Application()
        >>> ctx = chisel.Context(application, chisel.Context.create_environ('GET', '/hello'))
        >>> ctx.reconstruct_url()
        'http://localhost:80/hello'

        :param str path_info: Optional replacement for the URL path
        :param str query_string: Optional replacement for the query string
        :param bool relative: If True, creates a relative URL. Default is False.
        """

        environ = self.environ
        if relative:
            url = ''
        else:
            url = environ['wsgi.url_scheme'] + '://'

            if environ.get('HTTP_HOST'):
                url += environ['HTTP_HOST']
            else:
                url += environ['SERVER_NAME']

                if environ['wsgi.url_scheme'] == 'https':
                    if environ['SERVER_PORT'] != '443':
                        url += ':' + environ['SERVER_PORT']
                else:
                    if environ['SERVER_PORT'] != '80':
                        url += ':' + environ['SERVER_PORT']

        url += quote(environ.get('SCRIPT_NAME', ''))
        if path_info is None:
            url += quote(environ.get('PATH_INFO', ''))
        else:
            url += path_info
        if query_string is None:
            if environ.get('QUERY_STRING'):
                url += '?' + environ['QUERY_STRING']
        else:
            if query_string:
                if isinstance(query_string, str):
                    url += '?' + query_string
                else:
                    url += '?' + encode_query_string(query_string)

        return url


class StartResponse:
    """
    A WSGI start_response callable object that records its status and headers arguments

    >>> def application(environ, start_response):
    ...     start_response('200 OK', [('Content-Type', 'text/plain')])
    ...     return [b'Hello']
    >>> start_response = chisel.app.StartResponse()
    >>> application({}, start_response)
    >>> start_response.status, start_response.headers
    ('200 OK', [('Content-Type', 'text/plain')])

    :param status: The HTTP response status
    :type status: ~http.HTTPStatus or str
    :param list(tuple) headers: Optional list of key/value header tuples to add to the response
    """

    __slots__ = ('status', 'headers')

    def __init__(self):

        #: The start_response call's status argument
        self.status = None

        #: The start_response call's headers argument
        self.headers = None

    def __call__(self, status, headers):
        assert self.status is None and self.headers is None
        self.status = status
        self.headers = headers



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel encoding and decoding utilities
"""

from datetime import date, datetime, timezone
from decimal import Decimal
import json
from urllib.parse import quote, unquote
from uuid import UUID


class JSONEncoder(json.JSONEncoder):
    """
    A :class:`~json.JSONEncoder` sub-class with support for :class:`~datetime.datetime`, :class:`~datetime.date`,
    :class:`~decimal.Decimal`, and :class:`~uuid.UUID` objects.
    """

    __slots__ = ()

    def default(self, o):
        """
        The override of the :meth:`~json.JSONEncoder.default` method to add support for :class:`~datetime.datetime`,
        :class:`~datetime.date`, :class:`~decimal.Decimal`, and :class:`~uuid.UUID` objects.
        """

        if isinstance(o, datetime):
            return (o if o.tzinfo else o.replace(tzinfo=timezone.utc)).isoformat()
        if isinstance(o, date):
            return o.isoformat()
        if isinstance(o, Decimal):
            return float(o)
        if isinstance(o, UUID):
            return str(o)
        return json.JSONEncoder.default(self, o)


def encode_query_string(obj, encoding='utf-8'):
    """
    Encode an object as a query string. Dictionaries, lists, and tuples are recursed. Each member key is expressed in
    fully-qualified form. List keys are the index into the list, and are in order. For example:

    >>> import chisel
    >>> chisel.encode_query_string({'a': 5, 'b': 3.14, 'c': {'d': 'foo', 'e': [1, 2, 3]}, 'f': [{'g': True}, {'g': False}]})
    'a=5&b=3.14&c.d=foo&c.e.0=1&c.e.1=2&c.e.2=3&f.0.g=true&f.1.g=false'

    :param object obj: The object to encode as a query string
    :param str encoding: The query string encoding
    :returns: The encoded query string
    """

    return '&'.join(f'{k}={v}' for k, v in _encode_query_string_items(obj, None, encoding))

def _encode_query_string_items(obj, parent, encoding):
    if isinstance(obj, dict):
        if obj:
            for member, value in sorted(obj.items()):
                member_quoted = quote(str(member), encoding=encoding) if encoding else str(member)
                parent_member = f'{parent}.{member_quoted}' if parent else member_quoted
                yield from _encode_query_string_items(value, parent_member, encoding)
        elif parent:
            yield parent, ''
    elif isinstance(obj, (list, tuple)):
        if obj:
            for idx, value in enumerate(obj):
                parent_member = f'{parent}.{idx}' if parent else str(idx)
                yield from _encode_query_string_items(value, parent_member, encoding)
        elif parent:
            yield parent, ''
    else:
        if isinstance(obj, bool):
            yield parent, 'true' if obj else 'false' # quote safe
        elif isinstance(obj, int):
            yield parent, str(obj) # quote safe
        elif isinstance(obj, datetime):
            if not obj.tzinfo:
                obj = obj.replace(tzinfo=timezone.utc)
            yield parent, quote(obj.isoformat(), encoding=encoding) if encoding else obj.isoformat()
        elif isinstance(obj, date):
            yield parent, obj.isoformat() # quote safe
        elif isinstance(obj, UUID):
            yield parent, str(obj) # quote safe
        elif obj is None:
            yield parent, 'null' # quote safe
        else: # str, float
            yield parent, quote(str(obj), encoding=encoding) if encoding else str(obj)


def decode_query_string(query_string, encoding='utf-8'):
    """
    Decode an object from a query string. Each member key of the query string is expressed in fully-qualified
    form. List keys are the index into the list, must be in order. For example:

    >>> import chisel
    >>> chisel.decode_query_string('a=5&b=3.14&c.d=foo&c.e.0=1&c.e.1=2&c.e.2=3&f.0.g=true&f.1.g=false')
    {'a': '5', 'b': '3.14', 'c': {'d': 'foo', 'e': ['1', '2', '3']}, 'f': [{'g': 'true'}, {'g': 'false'}]}

    :param str query_string: The query string
    :param str encoding: The query string encoding
    :returns: The decoded object
    :raises ValueError: Query string is invalid
    """

    return _decode_query_string_items(
        (key_value_str.split('=') for key_value_str in query_string.split('&') if key_value_str),
        encoding
    )

def _decode_query_string_items(query_string_items, encoding):

    # Build the object
    result = [None]
    for key_value in query_string_items:

        # Split the key/value string
        try:
            key_str, value_str = key_value
            value = unquote(value_str, encoding=encoding) if encoding else value_str
        except ValueError:
            raise ValueError(f"Invalid key/value pair {'='.join(key_value)!r:.1000s}")

        # Find/create the object on which to set the value
        parent = result
        key_parent = 0
        for key in (unquote(key, encoding=encoding) if encoding else key for key in key_str.split('.')):
            obj = parent[key_parent]

            # Array key?  First "key" of an array must start with "0".
            if isinstance(obj, list) or (obj is None and key == '0'):

                # Create this key's container, if necessary
                if obj is None:
                    obj = parent[key_parent] = []

                # Create the index for this key
                try:
                    key = int(key)
                except:
                    raise ValueError(f"Invalid key/value pair {'='.join(key_value)!r:.1000s}")
                if key == len(obj):
                    obj.append(None)
                elif key < 0 or key > len(obj):
                    raise ValueError(f"Invalid key/value pair {'='.join(key_value)!r:.1000s}")

            # Dictionary key
            else:

                # Create this key's container, if necessary
                if obj is None:
                    obj = parent[key_parent] = {}

                # Create the index for this key
                if obj.get(key) is None:
                    obj[key] = None

            # Update the parent object and key
            parent = obj
            key_parent = key

        # Set the value
        if parent[key_parent] is not None:
            raise ValueError(f"Duplicate key {'='.join(key_value)!r:.1000s}")
        parent[key_parent] = value

    return result[0] if (result[0] is not None) else {}



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
The chisel package build script
"""

import re
import os

from setuptools import setup

MODULE_NAME = 'chisel'
PACKAGE_NAME = MODULE_NAME
TESTS_REQUIRE = []

def main():
    """
    The chisel package build script's main entry point
    """

    # Read the package version
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'src', MODULE_NAME, '__init__.py'), encoding='utf-8') as init_file:
        version = re.search(r"__version__ = '(.+?)'", init_file.read()).group(1)

    # Read the readme for use as the long description
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.rst'), encoding='utf-8') as readme_file:
        long_description = readme_file.read()

    # Do the setup
    setup(
        name=PACKAGE_NAME,
        description=('Light-weight WSGI application framework, schema-validated JSON APIs, and API documentation.'),
        long_description=long_description,
        long_description_content_type='text/x-rst',
        version=version,
        author='Craig Hobbs',
        author_email='craigahobbs@gmail.com',
        keywords='api json framework schema wsgi',
        url='https://github.com/craigahobbs/chisel',
        license='MIT',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Topic :: Internet :: WWW/HTTP :: WSGI',
            'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
            'Topic :: Software Development :: Libraries :: Application Frameworks'
        ],
        package_dir={
            '': 'src'
        },
        packages=[MODULE_NAME],
        package_data={
            '': [
                'static/*'
            ]
        },
        test_suite='tests',
        tests_require=TESTS_REQUIRE,
        extras_require={
            'tests': TESTS_REQUIRE
        }
    )

if __name__ == '__main__':
    main()



# Licensed under the MIT License
# https://github.com/craigahobbs/chisel/blob/master/LICENSE

"""
Chisel action class
"""

from cgi import parse_header
from functools import partial
from http import HTTPStatus
from json import loads as json_loads

from .app import Context
from .model import ValidationError, ValidationMode, TypeStruct, TYPE_STRING
from .request import Request
from .spec import SpecParser
from .util import decode_query_string


def action(action_callback=None, **kwargs):
    """
    Decorator for creating an :class:`~chisel.Action` object that wraps an action callback function. For example:

    >>> import chisel
    ...
    >>> @chisel.action(spec='''
    ... # Sum a list of numbers
    ... action sum_numbers
    ...     url
    ...         GET
    ...     query
    ...         # The list of numbers to sum
    ...         int[len > 0] numbers
    ...     output
    ...         # The sum of the numbers
    ...         int sum
    ... ''')
    ... def sum_numbers(ctx, req):
    ...    return {'sum': sum(req['numbers'])}
    ...
    >>> application = chisel.Application()
    >>> application.add_request(sum_numbers)
    >>> application.request('GET', '/sum_numbers', query_string='numbers.0=1&numbers.1=2&numbers.2=3')
    ('200 OK', [('Content-Type', 'application/json')], b'{"sum":6}')

    Chisel actions schema-validate their input before calling the callback function. For example:

    >>> status, _, response = application.request('GET', '/sum_numbers', query_string='numbers=1')
    >>> status
    '400 Bad Request'

    >>> import json
    >>> from pprint import pprint
    >>> pprint(json.loads(response.decode('utf-8')))
    {'error': 'InvalidInput',
     'member': 'numbers',
     'message': "Invalid value '1' (type 'str') for member 'numbers', expected "
                "type 'array' (query string)"}

    When :attr:`~chisel.Application.validate_output` the response dictionary is also validated to the output schema.

    :param ~collections.abc.Callable action_callback: The action callback function
    """

    if action_callback is None:
        return partial(action, **kwargs)
    return Action(action_callback, **kwargs).decorate_module(action_callback)


class ActionError(Exception):
    """
    An action error exception. Raise this exception within an action callback function to respond with an error.

    >>> import chisel
    ...
    >>> @chisel.action(spec='''
    ... action my_action
    ...     url
    ...         GET
    ...     errors
    ...         AlwaysError
    ... ''')
    ... def my_action(ctx, req):
    ...    raise chisel.ActionError('AlwaysError')
    ...
    >>> application = chisel.Application()
    >>> application.add_request(my_action)
    >>> application.request('GET', '/my_action')
    ('400 Bad Request', [('Content-Type', 'application/json')], b'{"error":"AlwaysError"}')

    :param str error: The error code
    :param str message: Optional error message
    :param status: The HTTP response status
    :type status: ~http.HTTPStatus or str
    """

    __slots__ = ('error', 'message', 'status')

    def __init__(self, error, message=None, status=None):
        super().__init__(error)

        #: The error code
        self.error = error

        #: The error message or None
        self.message = message

        #: The HTTP response status
        self.status = status


class _ActionErrorInternal(Exception):

    __slots__ = ('status', 'error', 'message', 'member')

    def __init__(self, status, error, message=None, member=None):
        super().__init__(error)
        self.status = status
        self.error = error
        self.message = message
        self.member = member


class Action(Request):
    """
    A schema-validated, JSON API request. An Action wraps a callback function that it calls when a request occurs. Here's
    an example of an action callback function:

    >>> def my_action(ctx, req):
    ...    return {}

    The first arugument, "ctx", is the :class:`~chisel.Context` object. The second argument is the request object which
    contiains the schema-validated, combined path parameters, query string parameters, and JSON request content
    parameters.

    :param ~collections.abc.Callable action_callback: The action callback function
    :param str name: The action request name
    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
        or None to match any. The second value is the URL path or None to use the default path.
    :param str doc: The documentation markdown text
    :param str doc_group: The documentation group
    :param ~chisel.SpecParser spec_parser: Optional specification parser to use for specification parsing
    :param str spec: Optional action specification (see :ref:`spec`). If a specification isn't provided it can be
        provided through the "spec_parser" argument.
    :param bool wsgi_response: If True, the callback function's response is a WSGI application function
        response. Default is False.
    :param str jsonp: Optional JSONP key
    """

    __slots__ = ('action_callback', 'model', 'wsgi_response', 'jsonp')

    def __init__(self, action_callback, name=None, urls=(('POST', None),), doc=None, doc_group=None,
                 spec_parser=None, spec=None, wsgi_response=False, jsonp=None):

        # Use the action model name, if available
        if name is None:
            name = action_callback.__name__

        # Spec provided?
        if spec_parser is None:
            spec_parser = SpecParser(spec=spec)
        elif spec is not None:
            spec_parser.parse_string(spec)
        model = spec_parser.actions.get(name)
        assert model is not None, f'Unknown action "{name}"'

        super().__init__(
            name=name,
            urls=model.urls or urls,
            doc=doc if doc is not None else model.doc,
            doc_group=doc_group if doc_group is not None else model.doc_group
        )

        #: The action callback function
        self.action_callback = action_callback

        #: The action's :class:`~chisel.ActionModel` object. This is retrieved from the spec/spec_parser
        #: attr:`~chisel.SpecParser.actions` collection.
        self.model = model

        #: If True, the callback function's response is a WSGI application function response.
        self.wsgi_response = wsgi_response

        #: JSONP key or None
        self.jsonp = jsonp

    def __call__(self, environ, unused_start_response):
        ctx = environ[Context.ENVIRON_CTX]

        # Handle the action
        is_get = (environ['REQUEST_METHOD'] == 'GET')
        jsonp = None
        validate_output = True
        try:
            # Read the request content
            try:
                content = None if is_get else environ['wsgi.input'].read()
            except:
                raise _ActionErrorInternal(HTTPStatus.REQUEST_TIMEOUT, 'IOError', message='Error reading request content')

            # De-serialize the JSON content
            try:
                if content:
                    content_type = environ.get('CONTENT_TYPE')
                    content_charset = ('utf-8' if content_type is None else parse_header(content_type)[1].get('charset', 'utf-8'))
                    content_json = content.decode(content_charset)
                    request = json_loads(content_json)
                else:
                    request = {}
            except Exception as exc:
                ctx.log.warning("Error decoding JSON content for action '%s'", self.name)
                raise _ActionErrorInternal(HTTPStatus.BAD_REQUEST, 'InvalidInput', message=f'Invalid request JSON: {exc}')

            # Validate the content
            try:
                request = self.model.input_type.validate(request, ValidationMode.JSON_INPUT)
            except ValidationError as exc:
                ctx.log.warning("Invalid content for action '%s': %s", self.name, str(exc))
                raise _ActionErrorInternal(
                    HTTPStatus.BAD_REQUEST,
                    'InvalidInput',
                    message=f'{exc} (content)',
                    member=exc.member
                )

            # Decode the query string
            query_string = environ.get('QUERY_STRING', '')
            try:
                request_query = decode_query_string(query_string)
            except Exception as exc:
                ctx.log.warning("Error decoding query string for action '%s': %.1000r", self.name, query_string)
                raise _ActionErrorInternal(HTTPStatus.BAD_REQUEST, 'InvalidInput', message=str(exc))

            # JSONP?
            if is_get and self.jsonp and self.jsonp in request_query:
                jsonp = str(request_query[self.jsonp])
                del request_query[self.jsonp]

            # Validate the query string
            try:
                request_query = self.model.query_type.validate(request_query, ValidationMode.QUERY_STRING)
            except ValidationError as exc:
                ctx.log.warning("Invalid query string for action '%s': %s", self.name, str(exc))
                raise _ActionErrorInternal(
                    HTTPStatus.BAD_REQUEST,
                    'InvalidInput',
                    message=f'{exc} (query string)',
                    member=exc.member
                )

            # Validate the path args
            request_path = ctx.url_args if ctx.url_args is not None else {}
            try:
                request_path = self.model.path_type.validate(request_path, ValidationMode.QUERY_STRING)
            except ValidationError as exc:
                ctx.log.warning("Invalid path for action '%s': %s", self.name, str(exc))
                raise _ActionErrorInternal(
                    HTTPStatus.BAD_REQUEST,
                    'InvalidInput',
                    message=f'{exc} (path)',
                    member=exc.member
                )

            # Copy top-level path keys and query string keys
            for request_key, request_value in request_path.items():
                request[request_key] = request_value
            for request_key, request_value in request_query.items():
                request[request_key] = request_value

            # Call the action callback
            try:
                status = HTTPStatus.OK
                response = self.action_callback(ctx, request)
                if self.wsgi_response:
                    return response
                if response is None:
                    response = {}
                response_type = self.model.output_type
            except ActionError as exc:
                status = exc.status or HTTPStatus.BAD_REQUEST
                response = {'error': exc.error}
                if exc.message is not None:
                    response['message'] = exc.message
                if ctx.app.validate_output:
                    if exc.error in ('UnexpectedError',):
                        validate_output = False
                    else:
                        response_type = TypeStruct()
                        response_type.add_member('error', self.model.error_type)
                        response_type.add_member('message', TYPE_STRING, optional=True)
            except Exception as exc:
                ctx.log.exception("Unexpected error in action '%s'", self.name)
                raise _ActionErrorInternal(HTTPStatus.INTERNAL_SERVER_ERROR, 'UnexpectedError')

            # Validate the response
            if validate_output and ctx.app.validate_output:
                try:
                    response_type.validate(response)
                except ValidationError as exc:
                    ctx.log.error("Invalid output returned from action '%s': %s", self.name, str(exc))
                    raise _ActionErrorInternal(HTTPStatus.INTERNAL_SERVER_ERROR, 'InvalidOutput', message=str(exc), member=exc.member)

        except _ActionErrorInternal as exc:
            status = exc.status
            response = {'error': exc.error}
            if exc.message is not None:
                response['message'] = exc.message
            if exc.member is not None:
                response['member'] = exc.member

        # Serialize the response as JSON
        return ctx.response_json(status, response, jsonp=jsonp)
